@@ -17,6 +17,7 @@ from sklearn.exceptions import ConvergenceWarning
 from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
 from sklearn.linear_model import Lasso, Ridge, ElasticNet, LogisticRegression
 from table_evaluator.helpers import *
+from dython.nominal import associations
 
 
 class TableEvaluator:
@@ -25,7 +26,7 @@ class TableEvaluator:
     Additional evaluations can be done with the different methods of evaluate and the visual evaluation method.
     """
 
-    def __init__(self, real: pd.DataFrame, fake: pd.DataFrame, cat_cols=None, unique_thresh=55, metric='pearsonr', verbose=False, n_samples=None,
+    def __init__(self, real: pd.DataFrame, fake: pd.DataFrame, cat_cols=None, unique_thresh=0, metric='pearsonr', verbose=False, n_samples=None,
                  name: str = None):
         """
 
@@ -41,8 +42,8 @@ class TableEvaluator:
         """
         self.name = name
         self.unique_thresh = unique_thresh
-        self.real = real
-        self.fake = fake
+        self.real = real.copy()
+        self.fake = fake.copy()
         self.comparison_metric = getattr(stats, metric)
         self.verbose = verbose
 
@@ -65,6 +66,11 @@ class TableEvaluator:
         self.fake = self.fake.sample(self.n_samples)
         assert len(self.real) == len(self.fake), f'len(real) != len(fake)'
 
+        self.real.loc[:, self.categorical_columns] = self.real.loc[:, self.categorical_columns].fillna('[NAN]')
+        self.fake.loc[:, self.categorical_columns] = self.fake.loc[:, self.categorical_columns].fillna('[NAN]')
+        self.real.loc[:, self.numerical_columns] = self.real.loc[:, self.numerical_columns].fillna(self.real[self.numerical_columns].mean()) 
+        self.fake.loc[:, self.numerical_columns] = self.fake.loc[:, self.numerical_columns].fillna(self.fake[self.numerical_columns].mean()) 
+
     def plot_mean_std(self):
         """
         Class wrapper function for plotting the mean and std using `plot_mean_std` from helpers.
@@ -161,18 +167,21 @@ class TableEvaluator:
         :param how: metric to measure distance. Choose from [``euclidean``, ``mae``, ``rmse``].
         :return: distance between the association matrices in the chosen evaluation metric. Default: Euclidean
         """
-        distance_func = None
         if how == 'euclidean':
             distance_func = euclidean_distance
         elif how == 'mae':
             distance_func = mean_absolute_error
         elif how == 'rmse':
             distance_func = rmse
+        elif how == 'cosine':
+            distance_func = cosine_similarity
+        else:
+            raise ValueError(f'`how` parameter must be in [euclidean, mae, rmse]')
 
-        assert distance_func is not None, f'Distance measure was None. Please select a measure from [euclidean, mae]'
-
-        real_corr = associations(self.real, cat_cols=self.categorical_columns, return_results=True, theil_u=True, plot=False)
-        fake_corr = associations(self.fake, cat_cols=self.categorical_columns, return_results=True, theil_u=True, plot=False)
+        fig, ax = plt.subplots()
+        real_corr = associations(self.real, nominal_columns=self.categorical_columns, theil_u=True, plot=False, ax=ax)['corr']
+        fake_corr = associations(self.fake, nominal_columns=self.categorical_columns, theil_u=True, plot=False, ax=ax)['corr']
+        plt.close(fig)
 
         return distance_func(
             real_corr.values,
@@ -183,8 +192,8 @@ class TableEvaluator:
         """
         Plot the first two components of a PCA of real and fake data.
         """
-        real = numerical_encoding(self.real, cat_cols=self.categorical_columns)
-        fake = numerical_encoding(self.fake, cat_cols=self.categorical_columns)
+        real = numerical_encoding(self.real, nominal_columns=self.categorical_columns)
+        fake = numerical_encoding(self.fake, nominal_columns=self.categorical_columns)
         pca_r = PCA(n_components=2)
         pca_f = PCA(n_components=2)
 
@@ -249,8 +258,8 @@ class TableEvaluator:
         real = self.real
         fake = self.fake
 
-        real = numerical_encoding(real, cat_cols=self.categorical_columns)
-        fake = numerical_encoding(fake, cat_cols=self.categorical_columns)
+        real = numerical_encoding(real, nominal_columns=self.categorical_columns)
+        fake = numerical_encoding(fake, nominal_columns=self.categorical_columns)
 
         self.pca_r.fit(real)
         self.pca_f.fit(fake)
@@ -368,7 +377,8 @@ class TableEvaluator:
         total_metrics = pd.DataFrame()
         for ds_name in ['real', 'fake']:
             ds = getattr(self, ds_name)
-            corr_df = associations(ds, cat_cols=self.categorical_columns, return_results=True, theil_u=True, plot=False)
+            corr_df = associations(ds, nominal_columns=self.categorical_columns, theil_u=True, plot=False)['corr']
+            plt.close()
             values = corr_df.values
             values = values[~np.eye(values.shape[0], dtype=bool)].reshape(values.shape[0], -1)
             total_metrics[ds_name] = values.flatten()
@@ -380,6 +390,7 @@ class TableEvaluator:
             print(total_metrics.to_string())
         return corr
 
+
     def convert_numerical(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
         """
         Special function to convert dataset to a numerical representations while making sure they have identical columns. This is sometimes a problem with
@@ -387,11 +398,11 @@ class TableEvaluator:
 
         :return: Real and fake dataframe with categorical columns one-hot encoded and binary columns factorized.
         """
-        real = numerical_encoding(self.real, cat_cols=self.categorical_columns)
+        real = numerical_encoding(self.real, nominal_columns=self.categorical_columns)
 
         columns = sorted(real.columns.tolist())
         real = real[columns]
-        fake = numerical_encoding(self.fake, cat_cols=self.categorical_columns)
+        fake = numerical_encoding(self.fake, nominal_columns=self.categorical_columns)
         for col in columns:
             if col not in fake.columns.tolist():
                 fake[col] = 0
@@ -413,11 +424,11 @@ class TableEvaluator:
         self.target_type = target_type
 
         # Convert both datasets to numerical representations and split x and  y
-        real_x = numerical_encoding(self.real.drop([target_col], axis=1), cat_cols=self.categorical_columns)
+        real_x = numerical_encoding(self.real.drop([target_col], axis=1), nominal_columns=self.categorical_columns)
 
         columns = sorted(real_x.columns.tolist())
         real_x = real_x[columns]
-        fake_x = numerical_encoding(self.fake.drop([target_col], axis=1), cat_cols=self.categorical_columns)
+        fake_x = numerical_encoding(self.fake.drop([target_col], axis=1), nominal_columns=self.categorical_columns)
         for col in columns:
             if col not in fake_x.columns.tolist():
                 fake_x[col] = 0
@@ -439,6 +450,7 @@ class TableEvaluator:
         self.real_x_train, self.real_x_test, self.real_y_train, self.real_y_test = train_test_split(real_x, real_y, test_size=0.2)
         self.fake_x_train, self.fake_x_test, self.fake_y_train, self.fake_y_test = train_test_split(fake_x, fake_y, test_size=0.2)
 
+
         if target_type == 'regr':
             self.estimators = [
                 RandomForestRegressor(n_estimators=20, max_depth=5, random_state=42),
@@ -484,8 +496,8 @@ class TableEvaluator:
         """
         if n_samples is None:
             n_samples = len(self.real)
-        real = numerical_encoding(self.real, cat_cols=self.categorical_columns)
-        fake = numerical_encoding(self.fake, cat_cols=self.categorical_columns)
+        real = numerical_encoding(self.real, nominal_columns=self.categorical_columns)
+        fake = numerical_encoding(self.fake, nominal_columns=self.categorical_columns)
 
         columns = sorted(real.columns.tolist())
         real = real[columns]
@@ -560,7 +572,7 @@ class TableEvaluator:
             'Correlation column correlations': correlation_correlation,
             'Mean Correlation between fake and real columns': column_correlation,
             f'{"1 - MAPE Estimator results" if self.target_type == "class" else "Correlation RMSE"}': estimators,
-            '1 - MAPE 5 PCA components': pca_variance,
+            # '1 - MAPE 5 PCA components': pca_variance,
         }
         total_result = np.mean(list(all_results.values()))
         all_results['Similarity Score'] = total_result
